#!/usr/bin/env ruby

##
# UPnPwn v0.8.0-bh
# Copyright 2008 Jonathan Squire, CISSP
##

require 'net/http'
require 'rexml/document'
require 'soap/rpc/driver'

require 'util'

####
# Implements an interface for accessing a UPnP Service
# general use:
#	s=UPnPService.new
#	s.controlURL=url
#	s.scpdURL=url
#	s.load_xml
#	s.parse_scpd
#	s.soap_init
#	s.the_action_you_want(args)
####
class UPnPService
	attr_accessor :serviceType, :serviceId, :controlURL, :eventSubURL, :scpdURL
	attr_accessor :xml_data

	#This will be used for the soap driver
	attr_accessor :soap

	
	#these will be hashes
	attr_accessor :actions
	#will be hash["action"]=>Array of paramaters
	#list of arguments to pass to an action
	attr_accessor :action_args_in
	#list of returns from an action
	attr_accessor :action_args_out
	
	def initialize(scpd_url=nil)
	
	  scpdURL=scpd_url
	  
	  if (!@scpdURL.nil?)
	    load_xml_data
	  end
	  
	  @actions = Array.new
	  @action_args_in = Hash.new
	  @action_args_out = Hash.new
	  
	end

	#
	# returns false if no data has been loaded into @xml_data
	#
	def xml?
		xml_loaded=false
		
		if(!@xml_data.nil?)
			xml_loaded = true
		end
		
		return xml_loaded
	end
	
	
	#
	# Retrives the SCPD description for the service and stores it in @xml_data
	#
	def load_xml_data
	  if (@scpdURL.nil?)
		puts "Error: No Location to load_xml_data from"
	  else
		@xml_data=Net::HTTP.get_response(URI.parse(@scpdURL)).body
	  end
	end
	
	#
	# returns a string with details about this instance
	#
	def info
		str = "UPnPService:\n"
		str = str + "Service Type: #{@serviceType}\n"
		str = str + "Service ID:   #{@serviceId}\n"
		str = str + "Control URL:  #{@controlURL}\n"
		str = str + "Event URL:    #{@eventSubURL}\n"
		str = str + "SCPD URL:     #{@scpdURL}\n"
		str = str + "# of Actions: #{@actions.size}\n"
		str = str + "Actions:      " + actions * ","
		return str
	end

	#
	# parses SCPD XML (from a string, by default, from @xml_data)
	# Fills in:
	#	@actions
	#	@action_args_in
	#	@action_args_out
	#
	def parse_scpd(doc=@xml_data)
		_doc=REXML::Document.new(doc)

		@actions = Array.new
		@action_args_in = Hash.new
		@action_args_out = Hash.new
		
		actions=_doc.elements.to_a("//action")

		actions.each {
			|action|
			tmp_doc=REXML::Document.new(action.to_s)
			#The spec had an extra space as a typo, so let's make sure there aren't any by using .strip
			cur_action=tmp_doc.text("//name").strip
			@actions << cur_action
			@action_args_in[cur_action]=Array.new
			@action_args_out[cur_action]=Array.new

			args=tmp_doc.elements.to_a("//argument")

			arg_in = Array.new
			arg_out = Array.new

			args.each {
				|arg|
				arg_doc=REXML::Document.new(arg.to_s)
				if (arg_doc.text("//direction")=="in")
					action_args_in[cur_action]<<arg_doc.text("//name").strip
				end
				if (arg_doc.text("//direction")=="out")
					action_args_out[cur_action]<<arg_doc.text("//name").strip
				end
			}
		}
	end

	#
	# returns a string containing info about a specific action
	#
	def action_info(action)
		ai = ""
		ai << "#########################\n"
		ai << "## #{action}\n"
		ai << "##\n"
		ai << "##      #{action}("
		if (@action_args_in[action].size > 0)
			ai << @action_args_in[action]  * ","
		end
		ai << ")\n"
		ai << "##\n"
		ai << "## Returns: " + @action_args_out[action] * "," + "\n"
		ai << "##\n"	

		return ai
	end
	
	#
	# prints out all actions
	#
	def print_actions
		
		actions.each {
			|action|

			puts action_info(action)
			puts add_method_code(action)
		}
	end

	#
	# helper function for building soap binding ruby file
	#
	def add_method_code(action)
		am = ""
		am << "meth_soapaction='##{action}'\n"
		am << "@soap.add_method_with_soapaction('#{action}',meth_namespace + meth_soapaction"
		if (@action_args_in[action].size > 0)
			am << ",'" + @action_args_in[action] * "','" + "'"
		end
		am << ")\n"
		
		return am
	end


	#
	# This function generates Ruby Code for dynamically adding a function via object.instance_eval
	#
	# Requires:
	#	action = "action from actions array"
	#
	#Optional:
	#	helper_method = false	#If false, we won't generate a helper method that forwards object.action to object.soap.action
	#
	def dynamic_action_code(action,helper_method=true)
		am = ""
		am << "@soap.add_method_with_soapaction('#{action}', @serviceType + '##{action}'"
		if (@action_args_in[action].size > 0)
			am << ",'" + @action_args_in[action] * "','" + "'"
		end
		am << ")\n"

		if(helper_method)
			am << "def #{action}(*args)\n"
			am << "  @soap.#{action}(*args)\n"
			am << "end\n"
		end
		
		return am
	end

	
	#
	# initializes soap driver and dynamically adds code to this instance to access each available service action
	#
	def soap_init
		@soap = SOAP::RPC::Driver.new(@controlURL,@serviceType)
		
		@actions.each {
			|action|
			self.instance_eval self.dynamic_action_code(action)
		}
	end
	
	
end

####
# Represents One UPnPHost
# used to store information such as it's IP, a count of all msgs, location url, etc
####
class UPnPHost
	attr_accessor :ip, :name, :msg_count, :last_msg, :location
	attr_accessor :xml_data
	attr_accessor :service
	
	def initialize()
	  @msg_count=0
	  @service=Hash.new
	end

	#
	# Return an array of all the services
	#
	def services
		array=Array.new
		
		
		@service.each {
			|key,value|
			array << key
		}
		
		return array
	end

	#
	# returns false if no data has been loaded into @xml_data
	#	
	def xml?
		xml_loaded=false
		
		if(!@xml_data.nil?)
			xml_loaded = true
		end
		
		return xml_loaded
	end
	
	#
	# returns a string containing details about this instance
	#
	def info
		str =  "UPnPHost:\n"
		str << "  IP:             #{@ip}\n"
		str << "  Location:           #{@location}\n"
		str << "  # of Services:  #{self.services.size}\n"
		str << "  Services:\n"
		str << "    #{self.services * "\n    "}\n"
		str << "  Total Messages: #{@msg_count}\n"
		str << "\n"
		
		if(!last_msg.nil?)
			str << "  Last Message (Size=#{@last_msg.size}):\n"
			str << "#{@last_msg}"
		end
		
		return str
	end
	
	def print
		puts "IP:             #{ip}"
		puts "Name:           #{name}"
		puts "Total Messages: #{msg_count}"
		puts "Last Message (Size=#{last_msg.size}):\n#{last_msg}"
	end
	
	def short_info
		return "#{@ip} MSGs=#{@msg_count} Location: #{@location}"
	end
	
  def load_xml_data
  
	if (@location.nil?)
		puts "Error: No Location to load_xml_data from"
	else
		@xml_data=Net::HTTP.get_response(URI.parse(@location)).body
	end
  end

  #
  # Refactor to determine type of message first
  #
  def parse_message(msg)
  	result=Hash.new
	
	split_msg=msg.split("\n")
	
	for line in split_msg
		#split at ':' at most two (2) parts
		key,value=line.split(':',2)
		#handle Method line (normally this only happens on the first line)
		if(value.nil?)
			value=key
			key="METHOD"
		end
		#convert to upper case key
		result[key.strip.upcase]=value.strip
	end
	
	return(result)
	
  end
  
  
  def save_msg(msg)
	@last_msg=msg
	
	msg_hash=parse_message(msg)

##REFACTOR
##This is pretty bad and can be easily overwritten	
	if(msg_hash["LOCATION"].nil?)
		#puts "NO LOCATION!"
		@location=nil
		else
		@location=msg_hash["LOCATION"]
		#puts "LOCATION: #{msg_hash["LOCATION"]}"
	end
	
  end

	#
	# parses device description (from @xml_data by default) for all services
	# Fills in:
	#	@service hash
	#
  def parse_services(doc=@xml_data)
	_doc=REXML::Document.new(doc)
	
	@service=Hash.new
	url=URI.parse(@location)
	url_prefix="#{url.scheme}://#{url.host}:#{url.port}"
	
	_services=_doc.elements.to_a("//device/deviceList/device/serviceList/service")
	_services.each {
		|svc|
		tmp_doc=REXML::Document.new(svc.to_s)
		cur_svc=tmp_doc.text("//serviceType")
		@service[cur_svc]=UPnPService.new
		@service[cur_svc].serviceType=tmp_doc.text("//serviceType")
		@service[cur_svc].serviceId=tmp_doc.text("//serviceId")
		@service[cur_svc].controlURL=url_prefix + tmp_doc.text("//controlURL")
		@service[cur_svc].eventSubURL=url_prefix + tmp_doc.text("//eventSubURL")
		@service[cur_svc].scpdURL=url_prefix + tmp_doc.text("//SCPDURL")
	}
	
  end
  
end


#
# Prints out all the port mappings
#
def list_portmaps(upnp_service)
	header = ""
	header << "      Source IP    External Port     Internal Host: Port [Enabled?, Lease, Desc]\n"
	header << "--------------------------------------------------------------------------------"
	puts header

	begin
		index=0
		while(true)
			result=upnp_service.GetGenericPortMappingEntry(index)
			index+=1
			puts portmap_to_s(result)
		end
		#If we get a UPnPError, we are done
		rescue => err
	end
end

#
# Expects 8 element Array
#
def portmap_to_s(ar=nil)
	p_str =""
	if (ar.length == 8)
		_NewRemoteHost=ar[0]
		_NewExternalPort=ar[1]
		_NewProtocol=ar[2]
		_NewInternalPort=ar[3]
		_NewInternalClient=ar[4]
		_NewEnabled=ar[5]
		_NewPortMappingDescription=ar[6]
		_NewLeaseDuration=ar[7]
		
		#
		# Lazy Hack, SOAP4R seems to return a SOAP::Mapping::Object if a field is empty
		#
		if(_NewRemoteHost.class!=String)
			_NewRemoteHost="0.0.0.0"
		end
		if(_NewInternalClient.class!=String)
			_NewInternalClient="0.0.0.0"
		end
	end
	
	p_str << "%15s" % "#{_NewRemoteHost}"
	p_str << " -> WAN:"
	p_str << "%5s" % "#{_NewExternalPort}"
	p_str << "/#{_NewProtocol} ->"
	p_str << "%15s" % "#{_NewInternalClient}"
	p_str << ":"
	p_str << "%5s" % "#{_NewInternalPort}"
	p_str << " [#{_NewEnabled}, #{_NewLeaseDuration}, #{_NewPortMappingDescription}]\n"
	
	return p_str
end



if __FILE__ == $0
  puts "Testing " + $0 + "\n" 
end